@@ -6,16 +6,16 @@ Example graph serialization using low-level Parquet read/write
 efficiently in Python.
 """
 
+import ast
 import csv
 import json
 import sys
 import typing
 
 from icecream import ic  # type: ignore  # pylint: disable=E0401
-from pydantic import BaseModel  # pylint: disable=E0401,E0611
+from pydantic import BaseModel, confloat, conint, NonNegativeInt, ValidationError  # pylint: disable=E0401,E0611
 from rich.progress import track  # pylint: disable=E0401
 import cloudpathlib
-import kglab
 import pandas as pd
 import pyarrow as pa  # type: ignore  # pylint: disable=E0401
 import pyarrow.lib  # type: ignore  # pylint: disable=E0401
@@ -27,9 +27,12 @@ import rdflib
 ## non-class definitions
 
 GraphRow = typing.Dict[str, typing.Any]
+IndexInts = conint(ge=-1)
 PropMap = typing.Dict[str, typing.Any]
+TruthType = confloat(ge=0.0, le=1.0)
 
-NOT_FOUND: int = -1
+EMPTY_STRING: str = ""
+NOT_FOUND: IndexInts = -1  # type: ignore
 
 
 ######################################################################
@@ -39,11 +42,11 @@ class Edge (BaseModel):  # pylint: disable=R0903
     """
 Representing an edge (arc) in the graph.
     """
-    BLANK_RELATION: typing.ClassVar[int] = 0
+    BLANK_RELATION: typing.ClassVar[NonNegativeInt] = 0
 
-    rel: int = BLANK_RELATION
-    node_id: int = NOT_FOUND
-    truth: float = 1.0
+    rel: NonNegativeInt = BLANK_RELATION
+    node_id: IndexInts = NOT_FOUND  # type: ignore
+    truth: TruthType = 1.0  # type: ignore
     prop_map: PropMap = {}
 
 
@@ -56,14 +59,14 @@ Representing a node (entity) in the graph.
     """
     BASED_LOCAL: typing.ClassVar[int] = -1
 
-    node_id: int = NOT_FOUND
-    name: str = ""
-    shadow: int = BASED_LOCAL
+    node_id: IndexInts = NOT_FOUND  # type: ignore
+    name: str = EMPTY_STRING
+    shadow: IndexInts = BASED_LOCAL  # type: ignore
     is_rdf: bool = False
     label_set: typing.Set[str] = set()
-    truth: float = 1.0
+    truth: TruthType = 1.0  # type: ignore
     prop_map: PropMap = {}
-    edge_map: typing.Dict[int, list] = {}
+    edge_map: typing.Dict[IndexInts, list] = {}  # type: ignore
 
 
     def add_edge (
@@ -88,12 +91,10 @@ class Partition (BaseModel):  # pylint: disable=R0903
     """
 Representing a partition in the graph.
     """
-    PROPS_NULL: typing.ClassVar[str] = "null"
-
-    part_id: int = NOT_FOUND
-    next_node: int = 0
-    nodes: typing.Dict[int, Node] = {}
-    node_names: typing.Dict[str, int] = {}
+    part_id: IndexInts = NOT_FOUND  # type: ignore
+    next_node: NonNegativeInt = 0
+    nodes: typing.Dict[NonNegativeInt, Node] = {}
+    node_names: typing.Dict[str, NonNegativeInt] = {}
     edge_rels: typing.List[str] = [""]
 
 
@@ -121,7 +122,7 @@ Lookup a node, return None if not found.
         """
 Create a name for a new node in the namespace, looking up first to avoid duplicates.
         """
-        node_id: int = NOT_FOUND
+        node_id: IndexInts = NOT_FOUND  # type: ignore
 
         if node_name in [None, ""]:
             raise ValueError(f"node name cannot be null |{ node_name }|")
@@ -136,7 +137,7 @@ Create a name for a new node in the namespace, looking up first to avoid duplica
 
 
     @classmethod
-    def load_props (
+    def _load_props (
         cls,
         props: str,
         *,
@@ -147,14 +148,14 @@ Load property pairs from a JSON string.
         """
         prop_map: PropMap = {}
 
-        if props not in (cls.PROPS_NULL, ""):
+        if props not in (EMPTY_STRING, "null"):
             prop_map = json.loads(props)
 
         return prop_map
 
 
     @classmethod
-    def save_props (
+    def _save_props (
         cls,
         prop_map: PropMap,
         *,
@@ -163,7 +164,7 @@ Load property pairs from a JSON string.
         """
 Save property pairs to a JSON string.
         """
-        props: str = cls.PROPS_NULL
+        props: str = EMPTY_STRING
 
         if len(prop_map) > 0:
             props = json.dumps(prop_map)
@@ -185,6 +186,27 @@ Add a node to the partition.
         self.nodes[node.node_id] = node
 
 
+    @classmethod
+    def _validation_error (
+        cls,
+        row_num: NonNegativeInt,
+        row: GraphRow,
+        message: str,
+        ) -> None:
+        """
+Print an error message to stderr.
+        """
+        print(
+            f"error at input row { row_num }: { message }",
+            file = sys.stderr,
+        )
+
+        print(
+            row,
+            file = sys.stderr,
+        )
+
+
     def populate_node (
         self,
         row: GraphRow,
@@ -194,20 +216,26 @@ Add a node to the partition.
         """
 Populate a Node object from the given Parquet row data.
         """
-        # create a src node
-        node: Node = Node(
-            node_id = self.create_node_name(row["src_name"]),
-            name = row["src_name"],
-            truth = row["truth"],
-            is_rdf = row["is_rdf"],
-            shadow = row["shadow"],
-            label_set = set(row["labels"].split(",")),
-            prop_map = self.load_props(row["props"]),
-        )
+        # first, lookup to make sure we don't overwrite if there are
+        # duplicates for the src node
+        src_name: str = row["src_name"]
+        src_node: typing.Optional[Node] = self.lookup_node(src_name, debug=debug)
 
-        self.add_node(node)
+        if src_node is None:
+            # create a src node
+            src_node = Node(
+                node_id = self.create_node_name(src_name, debug=debug),
+                name = row["src_name"],
+                truth = row["truth"],
+                is_rdf = row["is_rdf"],
+                shadow = row["shadow"],
+                label_set = set(row["labels"].split(",")),
+                prop_map = self._load_props(row["props"], debug=debug),
+            )
 
-        return node
+            self.add_node(src_node, debug=debug)  # type: ignore
+
+        return src_node  # type: ignore
 
 
     def get_edge_rel (
@@ -241,31 +269,68 @@ Populate an Edge object from the given Parquet row data.
         """
         # first, lookup the dst node and create if needed
         dst_name: str = row["dst_name"]
-        dst_node: typing.Optional[Node] = self.lookup_node(dst_name)
+        dst_node: typing.Optional[Node] = self.lookup_node(dst_name, debug=debug)
 
         if dst_node is None:
             dst_node = Node(
-                node_id = self.create_node_name(dst_name),
+                node_id = self.create_node_name(dst_name, debug=debug),
                 name = dst_name,
                 truth = row["truth"],
                 is_rdf = row["is_rdf"],
             )
 
-            self.add_node(dst_node)
+            self.add_node(dst_node, debug=debug)
 
         # create the edge
         edge: Edge = Edge(
-            rel = self.get_edge_rel(row["rel_name"], create=True),
+            rel = self.get_edge_rel(row["rel_name"], create=True, debug=debug),
             truth = row["truth"],
             node_id = dst_node.node_id,
-            prop_map = self.load_props(row["props"]),
+            prop_map = self._load_props(row["props"], debug=debug),
         )
 
-        node.add_edge(edge)
+        node.add_edge(edge, debug=debug)
 
         return edge
 
 
+    def dump_data (
+        self,
+        ) -> None:
+        """
+Dump the internal data structures for this partition.
+        """
+        for _, node_id in self.node_names.items():
+            node: Node = self.nodes[node_id]
+            ic(node)
+
+            for edge_rel, edge_list in node.edge_map.items():
+                for edge in edge_list:
+                    ic(edge_rel, edge)
+
+
+    @classmethod
+    def dump_parquet (
+        cls,
+        parq_file: pq.ParquetFile,
+        *,
+        debug: bool = False,
+        ) -> None:
+        """
+Dump the metadata and content for an input Parquet file.
+        """
+        ic(parq_file.metadata)
+        ic(parq_file.schema)
+        ic(parq_file.num_row_groups)
+
+        for batch in range(parq_file.num_row_groups):
+            row_group: pyarrow.lib.Table = parq_file.read_row_group(batch)  # pylint: disable=I1101
+
+            if row_group.num_rows > 0:
+                ic(row_group)
+                ic(row_group.columns)
+
+
     @classmethod
     def iter_load_parquet (
         cls,
@@ -276,7 +341,7 @@ Populate an Edge object from the given Parquet row data.
         """
 Iterate through the rows in a Parquet file.
         """
-        row_num: int = 0
+        row_num: NonNegativeInt = 0
 
         for batch in range(parq_file.num_row_groups):
             row_group: pyarrow.lib.Table = parq_file.read_row_group(batch)  # pylint: disable=I1101
@@ -292,7 +357,7 @@ Iterate through the rows in a Parquet file.
                         row[key] = val.as_py()
                     except IndexError as ex:
                         ic(ex, r_idx, c_idx)
-                        return
+                        sys.exit(-1)
 
                 if debug:
                     print()
@@ -306,26 +371,36 @@ Iterate through the rows in a Parquet file.
         self,
         csv_path: cloudpathlib.AnyPath,
         *,
+        encoding: str = "utf-8",
         debug: bool = False,
         ) -> typing.Iterable[typing.Tuple[int, GraphRow]]:
         """
 Iterate through the rows in a CSV file.
         """
-        row_num: int = 0
+        row_num: NonNegativeInt = 0
+
+        with open(csv_path, encoding=encoding) as fp:
+            reader = csv.reader(
+                fp,
+                delimiter = ",",
+                quotechar = '"',
+            )
 
-        with open(csv_path) as fp:
-            reader = csv.reader(fp, delimiter=",", quotechar='"')
             header = next(reader)
 
-            for row_val in reader:
-                row: GraphRow = dict(zip(header, row_val))
-                row["edge_id"] = int(row["edge_id"])
-                row["is_rdf"] = bool(row["is_rdf"])
-                row["shadow"] = int(row["shadow"])
-                row["truth"] = float(row["truth"])
+            try:
+                for row_val in reader:
+                    row: GraphRow = dict(zip(header, row_val))
+                    row["edge_id"] = int(row["edge_id"])
+                    row["is_rdf"] = bool(ast.literal_eval(row["is_rdf"]))
+                    row["shadow"] = int(row["shadow"])
+                    row["truth"] = float(row["truth"])
 
-                yield row_num, row
-                row_num += 1
+                    yield row_num, row
+                    row_num += 1
+            except ValueError as ex:
+                self._validation_error(row_num, row, str(ex))
+                sys.exit(-1)
 
 
     def iter_load_rdf (
@@ -333,31 +408,33 @@ Iterate through the rows in a CSV file.
         rdf_path: cloudpathlib.AnyPath,
         rdf_format: str,
         *,
+        encoding: str = "utf-8",
         debug: bool = False,
         ) -> typing.Iterable[typing.Tuple[int, GraphRow]]:
         """
 Iterate through the rows implied by a RDF file.
         """
-        row_num: int = 0
+        row_num: NonNegativeInt = 0
+        graph = rdflib.Graph()
 
-        kg = kglab.KnowledgeGraph()
-        kg.load_rdf(rdf_path, format=rdf_format)
-
-        for subj, pred, objt in kg.rdf_graph():
-            if debug:
-                ic(subj, pred, objt)
+        graph.parse(
+            rdf_path,
+            format = rdf_format,
+            encoding = encoding,
+        )
 
+        for subj in graph.subjects(unique=True):  # type: ignore
             # node representation for a triple
             row: GraphRow = {}
             row["src_name"] = str(subj)
             row["truth"] = 1.0
             row["edge_id"] = NOT_FOUND
-            row["rel_name"] = None
-            row["dst_name"] = None
+            row["rel_name"] = EMPTY_STRING
+            row["dst_name"] = EMPTY_STRING
             row["is_rdf"] = True
             row["shadow"] = Node.BASED_LOCAL
-            row["labels"] = ""
-            row["props"] = self.PROPS_NULL
+            row["labels"] = EMPTY_STRING
+            row["props"] = EMPTY_STRING
 
             if debug:
                 ic("node", subj, row_num, row)
@@ -365,23 +442,27 @@ Iterate through the rows implied by a RDF file.
             yield row_num, row
             row_num += 1
 
-            # edge representation for a triple
-            row = {}
-            row["src_name"] = str(subj)
-            row["truth"] = 1.0
-            row["edge_id"] = 1
-            row["rel_name"] = str(pred)
-            row["dst_name"] = str(objt)
-            row["is_rdf"] = True
-            row["shadow"] = Node.BASED_LOCAL
-            row["labels"] = ""
-            row["props"] = self.PROPS_NULL
+            for _, pred, objt in graph.triples((subj, None, None)):
+                if debug:
+                    ic(subj, pred, objt)
 
-            if debug:
-                ic("edge", objt, row_num, row)
+                # edge representation for a triple
+                row = {}
+                row["src_name"] = str(subj)
+                row["truth"] = 1.0
+                row["edge_id"] = 1
+                row["rel_name"] = str(pred)
+                row["dst_name"] = str(objt)
+                row["is_rdf"] = True
+                row["shadow"] = Node.BASED_LOCAL
+                row["labels"] = EMPTY_STRING
+                row["props"] = EMPTY_STRING
 
-            yield row_num, row
-            row_num += 1
+                if debug:
+                    ic("edge", objt, row_num, row)
+
+                yield row_num, row
+                row_num += 1
 
 
     def parse_rows (
@@ -396,11 +477,15 @@ Parse a stream of rows to construct a graph partition.
         for row_num, row in track(iter_load, description=f"parse rows"):
             # have we reached a row which begins a new node?
             if row["edge_id"] < 0:
-                node: Node = self.populate_node(row)
+                try:
+                    node: Node = self.populate_node(row, debug=debug)
 
-                if debug:
-                    print()
-                    ic(node)
+                    if debug:
+                        print()
+                        ic(node)
+                except ValidationError as ex:
+                    self._validation_error(row_num, row, str(ex))
+                    sys.exit(-1)
 
             # validate the node/edge sequencing and consistency among the rows
             elif row["src_name"] != node.name:
@@ -410,20 +495,38 @@ Parse a stream of rows to construct a graph partition.
 
             # otherwise this row is an edge for the most recent node
             else:
-                edge: Edge = self.populate_edge(row, node)
+                try:
+                    edge: Edge = self.populate_edge(row, node, debug=debug)
 
-                if debug:
-                    ic(edge)
+                    if debug:
+                        ic(edge)
+                except ValidationError as ex:
+                    self._validation_error(row_num, row, str(ex))
+                    sys.exit(-1)
 
 
     def iter_gen_rows (
         self,
+        *,
+        sort: bool = False,
+        debug: bool = False,
         ) -> typing.Iterable[GraphRow]:
         """
 Iterator for generating rows on writes.
+
+Optionally, sort on:
+  * src `node.name` in ASC order
+  * `edge_id` and dst `node.name` in ASC order
         """
-        for node in self.nodes.values():
-            yield {
+        if sort:
+            node_iter = sorted(self.node_names.items())
+        else:
+            node_iter = self.node_names.items()  # type: ignore
+
+        for _, node_id in node_iter:
+            node: Node = self.nodes[node_id]
+
+            row = {
                 "src_name": node.name,
                 "edge_id": -1,
                 "rel_name": None,
@@ -432,14 +535,26 @@ Iterator for generating rows on writes.
                 "shadow": node.shadow,
                 "is_rdf": node.is_rdf,
                 "labels": ",".join(node.label_set),
-                "props": self.save_props(node.prop_map),
+                "props": self._save_props(node.prop_map, debug=debug),
             }
 
-            edge_id: int = 0
+            yield row
 
-            for _, edge_list in node.edge_map.items():
-                for edge in edge_list:
-                    yield {
+            edge_id: NonNegativeInt = 0
+
+            if sort:
+                edge_rel_iter = sorted(node.edge_map.items())
+            else:
+                edge_rel_iter = node.edge_map.items()  # type: ignore
+
+            for _, edge_list in edge_rel_iter:
+                if sort:
+                    edge_iter = sorted(edge_list, key=lambda e: self.nodes[e.node_id].name)
+                else:
+                    edge_iter = edge_list
+
+                for edge in edge_iter:
+                    row = {
                         "src_name": node.name,
                         "edge_id": edge_id,
                         "rel_name": self.edge_rels[edge.rel],
@@ -448,31 +563,48 @@ Iterator for generating rows on writes.
                         "shadow": -1,
                         "is_rdf": node.is_rdf,
                         "labels": None,
-                        "props": self.save_props(edge.prop_map),
+                        "props": self._save_props(edge.prop_map, debug=debug),
                     }
 
+                    yield row
                     edge_id += 1
 
 
     def to_df (
         self,
+        *,
+        sort: bool = False,
+        debug: bool = False,
         ) -> pd.DataFrame:
         """
 Represent the partition as a DataFrame.
         """
-        return pd.DataFrame([row for row in self.iter_gen_rows()])
+        return pd.DataFrame([
+            row
+            for row in self.iter_gen_rows(
+                    sort = sort,
+                    debug = debug,
+            )
+        ])
 
 
     def save_file_parquet (
         self,
         save_parq: cloudpathlib.AnyPath,
         *,
+        sort: bool = False,
         debug: bool = False,
         ) -> None:
         """
 Save a partition to a Parquet file.
         """
-        table = pa.Table.from_pandas(self.to_df())
+        table = pa.Table.from_pandas(
+            self.to_df(
+                sort = sort,
+                debug = debug,
+            ),
+        )
+
         writer = pq.ParquetWriter(save_parq.as_posix(), table.schema)
         writer.write_table(table)
         writer.close()
@@ -482,38 +614,60 @@ Save a partition to a Parquet file.
         self,
         save_csv: cloudpathlib.AnyPath,
         *,
+        encoding: str = "utf-8",
+        sort: bool = False,
         debug: bool = False,
         ) -> None:
         """
 Save a partition to a CSV file.
         """
-        self.to_df().to_csv(save_csv.as_posix(), index=False)
+        self.to_df(
+            sort = sort,
+            debug = debug,
+        ).to_csv(
+            save_csv.as_posix(),
+            index = False,
+            header = True,
+            encoding = encoding,
+            quoting = csv.QUOTE_NONNUMERIC,
+        )
 
 
     def save_file_rdf (
         self,
         save_rdf: cloudpathlib.AnyPath,
-        rdf_format: str,
         *,
+        rdf_format: str = "ttl",
+        encoding: str = "utf-8",
+        sort: bool = False,
         debug: bool = False,
         ) -> None:
         """
 Save a partition to an RDF file.
         """
-        kg = kglab.KnowledgeGraph()
         subj = None
+        graph = rdflib.Graph()
 
-        for row in self.iter_gen_rows():
+        row_iter = self.iter_gen_rows(
+            sort = sort,
+            debug = debug,
+        )
+
+        for row in row_iter:
             if row["is_rdf"]:
-
                 if row["edge_id"] < 0:
                     subj = rdflib.term.URIRef(row["src_name"])
                 else:
                     pred = rdflib.term.URIRef(row["rel_name"])
                     objt = rdflib.term.URIRef(row["dst_name"])
-                    kg.add(subj, pred, objt)
+                    
+                    graph.add((subj, pred, objt))  # type: ignore
 
                     if debug:
                         ic(subj, pred, objt)
 
-        kg.save_rdf(save_rdf, format=rdf_format)
+        graph.serialize(
+            save_rdf,
+            format = rdf_format,
+            encoding = encoding,
+        )
